/*******************************
 * attack.c
 ********************************

 Author = Dr. Frank Schuhmacher 

	professional Security Analyst and Consultant for

	 	- Security Hardware Design 
		- Security Protocols
		- Security Software Design
		- Product Certification
		- Training
	
	with seven years experience in the Security Analysis of

		- Smart Cards
		- Systems on Chips
		- Embedded Systems
		- Cryptographic Protocols 

	If you want to encance the Security Level of your product, feel free to contact me.

 Phone = 0049-228-9762-9162
 email = frank.schuhmacher@segrids.com		
 Date = 2014-02-26
 Licence = GPLv3
 Acknowledgements = 

	This program includes software developed by the OpenSSL Project
     	for use in the OpenSSL Toolkit (http://www.openssl.org/).

	This program includes software developed by the GSL Project.
 
*********************************/


/************* includes **********/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <stdint.h>
#include <unistd.h>
#include <openssl/aes.h>
#include <math.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_sort.h>
#include <gsl/gsl_sort_vector.h>

/*********** defines *******************/
#define TRACELEN 435002
#define NPC 10 //number of principal components to be considered
#define NN 393 //length of region of interest for each s-box

/*********** roi (region of interest structure and type ************/
struct roi_struct {
	uint32_t N;
	uint32_t *vec; //Vector of trace offsets
	//redundant but practical: devide roi in intervals
	uint32_t n_intervals;
	uint32_t *offset;	
	uint32_t *length;	
};

typedef struct roi_struct roi_t;

/*********** global variables **************************/
unsigned int comb[16] = {98986,137079,103748,141841,108510,146602,113271,151364,118033,156125,122794,160887,127556,165649,132318,170411};

uint8_t ttt_mask[16]= {0x00, 0x0f, 0x36, 0x39, 0x53, 0x5c, 0x65, 0x6a, 0x95, 0x9a, 0xa3, 0xac, 0xc6, 0xc9, 0xf0, 0xff};

uint8_t testkey[32] = {0x6c, 0xec, 0xc6, 0x7f, 0x28, 0x7d, 0x08, 0x3d, 0xeb, 0x87, 0x66, 0xf0, 0x73, 0x8b, 0x36, 0xcf, 
	0x16, 0x4e, 0xd9, 0xb2, 0x46, 0x95, 0x10, 0x90, 0x86, 0x9d, 0x08, 0x28, 0x5d, 0x2e, 0x19, 0x3b};
uint8_t secondkeyhalf[16] ={0x16, 0x4e, 0xd9, 0xb2, 0x46, 0x95, 0x10, 0x90, 0x86, 0x9d, 0x08, 0x28, 0x5d, 0x2e, 0x19, 0x3b};

roi_t roi_j, roi_k; //roi_k contains offsets from comb
gsl_matrix *Trafo_j;
gsl_matrix *Templ_j;
gsl_matrix *Trafo_k;
gsl_matrix *Templ_k;

FILE *fifo_from_wrapper;
FILE *fifo_to_wrapper;

int8_t dummy_measure[435035]; //plain(16) | cipher(16) | '\0' | trace(435002)

/************* general functions ***************/
void print64(uint64_t b){
        printf("%016llX", (long long int)b);
}

/************ roi (region of interest) functions ***************/

int roi_load(roi_t *roi, char *roi_filename){
	roi->n_intervals = 0;
	FILE *f;
	roi->N = 0;
	roi->n_intervals = 0;
	// allocate enough memory, to be reduced below
	roi->vec = malloc(4*TRACELEN);
	roi->offset = malloc(4*TRACELEN);
	roi->length = malloc(4*TRACELEN);
	// open file
	printf("Loading %s\n", roi_filename);
	f = fopen(roi_filename, "r");
	if (f == NULL){
		printf("Error in roi_load: can't open %s\n", roi_filename);
		return -1;
	}
	// read first point into buffer:
	if (fscanf(f,"%d\n", roi->vec) == EOF){
		printf("Error in roi_load: File empty");
		return -1;
	} else {
		roi->offset[0] =roi->vec[0];
		roi->N = 1;
		roi->n_intervals = 1;
		roi->length[0]=1;
	}
	//read all other points into buffer
	while (fscanf(f,"%d\n", &roi->vec[roi->N]) != EOF){
		if (roi->vec[roi->N] > TRACELEN-1){
			printf("Error in roi_load: %d out of range", roi->vec[roi->N]);
			return -1;
		}
		if (roi->vec[roi->N] <= roi->vec[roi->N -1]){
			printf("Error in roi_load: %d <= %d", roi->vec[roi->N], roi->vec[roi->N -1]);
			return -1;
		}
		if (roi->vec[roi->N] - roi->vec[roi->N -1] == 1){
			roi->length[roi->n_intervals-1]++;
		} else {
			roi->offset[roi->n_intervals]=roi->vec[roi->N];
			roi->length[roi->n_intervals]=1;
			roi->n_intervals++;
		}
		roi->N++;
	}
	fclose(f);
	roi->vec = realloc(roi->vec, 4*roi->N);
	roi->offset = realloc(roi->offset, 4*roi->n_intervals);
	roi->length = realloc(roi->length, 4*roi->n_intervals);
	return 0;
}


/************* communcation functions ***************/
uint32_t start_talking(){ 
	//open channel
	fifo_from_wrapper = fopen("../fifo_from_wrapper", "r");
	fifo_to_wrapper = fopen("../fifo_to_wrapper", "w");
	setvbuf(fifo_to_wrapper, NULL, _IONBF, 0);
	setvbuf(fifo_from_wrapper, NULL, _IONBF, 0);
	uint32_t M;
	int ret;
	// receive M (= number of traces the wrapper will send) in 4 bytes 
	ret =fread(&M, 1, 4, fifo_from_wrapper);
	if (ret != 4){
		printf("Error in start talking: only %d bytes received\n", ret);
		return -1;		
	}
	printf("start_talking: M  = %u\n", M);
	fwrite((char*) "\x0a\x2e\x0a", 1, 3, fifo_to_wrapper);
	return M; 
}

int read_measure_from_fifo(uint8_t *plain, uint8_t *cipher, int8_t *trace){
	//setvbuf(stdin, NULL, _IONBF, 0);
	//receive plain
	int ret;
	ret =fread((void *)plain, 1, 16, fifo_from_wrapper);
	if (ret < 16){
		printf("Error in read_measure_from_stdin: only %d byte plaintext received\n", ret);
		return -1;
	}
//	printf("plain = ");
//	print64(((uint64_t *)plain)[0]);
//	print64(((uint64_t *)plain)[1]);
	//receive cipher
	ret =fread((void *)cipher, 1, 16, fifo_from_wrapper);
	if (ret < 16){
		printf("Error in read_measure_from_stdin: only %d byte ciphertext received\n", ret);
		return -1;
	}
//	printf("\ncipher = ");
//	print64(((uint64_t *)cipher)[0]);
//	print64(((uint64_t *)cipher)[1]);
//	printf("\n");
	//receive mask offset dummy
	uint8_t j_dummy= 0x69;
	fread((void *)&j_dummy, 1, 1, fifo_from_wrapper);
	if (j_dummy != 0){
		printf("Error in read_measure_from_stdin: mask offset not zero\n");
		return -1;
	}
//	printf("Dummy OK\n");
	//receive trace
	ret = fread((void *)trace, 1, 435002, fifo_from_wrapper);
	if (ret < 435002){
		printf("Error in read_measure_from_stdin: only %d bytes of trace received\n", ret);
		return -1;
	}
//	printf("Trace received\n");
	return 0;
}


void send_result_to_fifo(gsl_matrix_uchar *KK){
	// 1 unsigned byte '\0' (i.e. attack subkey used during the first round of the AES)
	fwrite((void*)"\0", 1, 1, fifo_to_wrapper);
	gsl_matrix_uchar_fwrite(fifo_to_wrapper, KK);
}

void handle_success(uint8_t *keyhalf, unsigned int m){
	unsigned int sbox, k; 
	uint8_t kd;
//	printf("The attack needed %d trace(s)\n", m);
//	print64(((uint64_t *)keyhalf)[0]);
//	print64(((uint64_t *)keyhalf)[1]);
//	printf("\n");
	while(1){
		fwrite((void*)"\0", 1, 1, fifo_to_wrapper);
		for (sbox=0; sbox < 16; sbox++){
			kd = keyhalf[sbox];
			for (k=0; k<256; k++){
				fwrite(&kd, 1,1, fifo_to_wrapper);	
			}
		}
		if (fread((void *)dummy_measure, 1, 435035, fifo_from_wrapper) < 435035){
			printf("no more data\n");
			return;
		}
	}
}

int init_templates(){
	Trafo_j = gsl_matrix_alloc(705, 10);
	Templ_j = gsl_matrix_alloc(16, 10);
	Trafo_k = gsl_matrix_alloc(256*NN, NPC);
	Templ_k = gsl_matrix_alloc(256*256, NPC);
	FILE *f;
	f = fopen("templates/Trafo_j_f.gsl_matrix", "r");
	if (gsl_matrix_fread(f, Trafo_j)){
		printf("Error: Unable to read Trafo_j from file.\n");
		return -1;
	}
	fclose(f);
	f = fopen("templates/Templ_j_f.gsl_matrix", "r");
	if (gsl_matrix_fread(f, Templ_j)){
		printf("Error: Unable to read Templ_j from file.\n");
		return -1;
	}
	fclose(f);
	f = fopen("templates/Trafo_misx_c.gsl_matrix", "r");
	if (gsl_matrix_fread(f, Trafo_k)){
		printf("Error: Unable to read Trafo_k from file.\n");
		return -1;
	}
	fclose(f);
	f = fopen("templates/Templ_misx_c.gsl_matrix", "r");
	if (gsl_matrix_fread(f, Templ_k)){
		printf("Error: Unable to read Templ_misx from file.\n");
		return -1;
	}
	fclose(f);
	return 0;
}


int init_roi(){
	if (roi_load(&roi_j, "templates/roi_j_f.txt")){
		printf("Error: roi_load failed for mskofs\n");
		return -1;
	}
	//roi_print(&roi_j);
	if (roi_load(&roi_k, "templates/roi_misx_c.txt")){
		printf("Error: roi_load failed for misx\n");
		return -1;
	}
	//roi_print(&roi_k);
	return 0;
}


int rank_j( gsl_permutation *jj, gsl_vector *u){
	unsigned int j, l, n;
	gsl_vector *dsq = gsl_vector_calloc(16);
	//dsq_j(dd_j, u);	
	double tru_n; 	// (Trafo_j * u)[n]
	double *Trafo_col_pointer;
	double *Templ_row_pointer; //c-interpretatino of "row" and "col"
	for (j=0; j<16; j++){
		//dd_j[j] = sum((tru-Templ_mskofs[,j])**2)
		Templ_row_pointer = Templ_j->data + 10*j;
		for (n=0; n<10; n++){
			tru_n = 0;
			Trafo_col_pointer = Trafo_j->data + n;
			for (l=0; l<705; l++){
				tru_n = tru_n + (Trafo_col_pointer[10*l] * u->data[l]);
			}
			dsq->data[j] = dsq->data[j] + ((tru_n - Templ_row_pointer[n])*(tru_n - Templ_row_pointer[n]));
		}
	}
	gsl_sort_vector_index(jj, dsq);
	return 0;
}

int dsq_k(gsl_matrix *dd, const uint8_t j, const uint8_t *plain, const gsl_matrix *uu){
	gsl_matrix_set_zero(dd);
	uint8_t sbox, idx;
	for (sbox=0; sbox<16; sbox++){
		idx = (j+sbox)%16;
		uint8_t mp = ttt_mask[idx]^plain[sbox]; //masked plaintext
		unsigned int i = 16*sbox + idx;
		//The Template depends on i
		double *Trafo = Trafo_k->data + i*NPC*NN;
		double *Templ = Templ_k->data + i*256*NPC;
		double Trafo_subtrace_n; //actual coefficient of Trafo(subtrace)
		unsigned int l, n;
		double *Trafo_col_pointer, *Templ_col_pointer; //c-meaning of "row" and "col"
		unsigned int subkey;
		uint8_t pkm;
		for (n=0; n < NPC; n++){
			Trafo_col_pointer = Trafo + n;
			Templ_col_pointer = Templ + n;
			Trafo_subtrace_n = 0;
			for (l=0; l<NN; l++){
				Trafo_subtrace_n = Trafo_subtrace_n + (Trafo_col_pointer[NPC*l] * uu->data[sbox*NN + l]);
			}
			//printf("Trafo_subtrace_n = %g\n", Trafo_subtrace_n);
			for (subkey=0; subkey<256; subkey++){
				pkm = mp^(uint8_t)subkey; //hypothetical masked sbox input		
			dd->data[sbox*256 + subkey] = dd->data[sbox*256 + subkey] +(Trafo_subtrace_n - Templ_col_pointer[pkm*NPC])*(Trafo_subtrace_n - Templ_col_pointer[pkm*NPC]);
			}
		}
	}
	return 0;			
}


void KK_by_dd(gsl_matrix_uchar *KK, gsl_matrix *dd){
	gsl_permutation *kk = gsl_permutation_alloc(256);
	gsl_vector_view dd_row;
	unsigned int sbox, i;
	for (sbox=0; sbox <16; sbox++){
		dd_row = gsl_matrix_row(dd, sbox); 
		gsl_sort_vector_index(kk, &dd_row.vector);
		//printf("key ranking for sbox %d: ", sbox);
		//gsl_permutation_fprintf (stdout, kk, " %u");
		for (i=0; i<256; i++){
			KK->data[sbox*256 + i] = (unsigned char)kk->data[i];
		}
	}
}

				
int key_check(const uint8_t *plain, const uint8_t *cipher, const uint8_t *keyguess){
	AES_KEY aes_key;
        unsigned char *out=malloc(16);
        AES_set_encrypt_key((const unsigned char *)keyguess, 256, &aes_key);
        AES_encrypt((const unsigned char *)plain, out, &aes_key);
//        int i;
	if ( memcmp(out, cipher, 16) ){
//                printf("keyguess wrong: ");
//                for(i=0; i<16; i++){
//			printf("%u ", keyguess[i]);
//		}
//		printf("\n");
		return 1;
        }
	return 0;
}	

void suboptimal(uint8_t *userank, int8_t r, int8_t s){
	uint8_t sbox;
	for (sbox=0; sbox <16; sbox++){
		userank[sbox] = 0;
	}
	if (r >= 0){
		userank[r]++;
	}
	if (s >= 0){
		userank[s]++;
	}
	
}

int key_by_KK(uint8_t *keyhalf, const gsl_matrix_uchar *KK, const uint8_t *plain, const uint8_t *cipher){
	//16 Bytes of memory for userank must already be allocated in calling function
	uint8_t keyguess[32];
	memcpy(keyguess+16, secondkeyhalf, 16);
	uint8_t sbox;
	uint8_t userank[16];
	int8_t r, s;
	for (r=-1; r<16; r++){
		for (s=-1; s<=r; s++){
			suboptimal(userank, r, s);
			for (sbox=0; sbox<16; sbox++){
				keyguess[sbox] = (uint8_t)KK->data[256*sbox + userank[sbox]]; //most propable subkey for sbox i
			}	
			if ( key_check(plain, cipher, (const uint8_t *)keyguess) == 0){
				if ((r>=0)|(s>=0)){
					printf("key found (r,s)=(%d, %d)\n",r,s);
				}
				//alt: keyhalf = (uint8_t *)keyguess;
				memcpy(keyhalf, keyguess, 16);
				return 1; //key found
			}
		}
	}
	return 0; //was unable to find key
}


int main(int argc, char *argv[]) {
	char str[] = "xx";
	unsigned int l, m, n, sbox;
	unsigned int do_check = 0;
	if (argc > 1){
		do_check = 1;
		if( strlen(argv[1]) != 32 ){
			printf("argument %s has wrong length %u\n", (char *)argv[1], (unsigned int)strlen(argv[1]));
			return -1;
		}
		for (l=0; l<16; l++){
			memcpy(str, argv[1]+(2*l), 2);
			//printf("secondkeyhalf[%d] = %02x\n", l, (unsigned char)strtol(str,NULL,16));
			secondkeyhalf[l] = (uint8_t)strtol(str,NULL,16);	
		}
		printf("thank you for the hint!\n");
	}
	init_roi();
	init_templates();
	uint8_t *plain = malloc(16);
	uint8_t *cipher = malloc(16);
	uint8_t j; 		//actual fovorite j in jj[0], jj[1], ...
	int8_t *trace = malloc(435002);
	int8_t *ptr;
	gsl_permutation *jj = gsl_permutation_alloc(16);
	gsl_matrix_uchar *KK = gsl_matrix_uchar_alloc(16,256); //for j[0]
	gsl_matrix *dd = gsl_matrix_calloc(16,256); 
	gsl_matrix *dd_best = gsl_matrix_calloc(16,256); 
	gsl_matrix *dd_sum = gsl_matrix_calloc(16,256); //sum of dd for more than one measure m
	uint8_t *keyhalf = malloc(16);
	uint32_t M;
	M = start_talking();
	for(m=0; m<M; m++){ 
		//get measure data
		read_measure_from_fifo( plain, cipher, trace);	
		//write mask offset specific subtrace into vector u
		gsl_vector *u = gsl_vector_alloc(705);
		for (n=0; n<705; n++){
			u->data[n] = (double)trace[roi_j.vec[n]];  
		}
		//rank possible offsets
		rank_j(jj, u);
		//write 16 sbox specific subtraces into matrix uu
		gsl_matrix *uu = gsl_matrix_alloc(16, NN);
		for (sbox=0; sbox<16; sbox++){
			ptr = (int8_t *)trace + comb[sbox]; //points to start of sbox
			for (n=0; n<NN; n++){
				uu->data[sbox*NN + n] = (double)ptr[roi_k.vec[n]];  
			}
		}
		//handle best j candidate jj[0] first 
		//j=jj[0] specific distance square matrix dd
		dsq_k(dd_best, jj->data[0], plain, uu);
		if (do_check == 0) goto POSTCHECK1;
		KK_by_dd(KK, dd_best);
		if (key_by_KK(keyhalf, KK, plain, cipher) == 1){
			handle_success(keyhalf, m);
			printf("good bye\n");
			return 0;
		}
		for (l=1; l<16; l++){ //jj[1],...,jj[s],...,jj[15]
			j = jj->data[l];
			// calculate j specific distance squared matrix dd
			dsq_k(dd, j, plain, uu);
			// and the corresponding key guess matrix KK
			KK_by_dd(KK, dd);	
			//search for keyhalf at prominent ranks in KK
			if (key_by_KK(keyhalf, KK, plain, cipher) == 1){
				handle_success(keyhalf, m);
				printf("good bye\n");
				return 0;
			}
		}
		//we are here because we were not able to disclose the key with a single trace 
		//we will try dd_sum to search for the key
		POSTCHECK1:
		gsl_matrix_add( dd_sum, dd_best); 
		KK_by_dd(KK, dd_sum);
		if (do_check == 0) goto POSTCHECK2;
		if (key_by_KK(keyhalf, KK, plain, cipher) == 1){
			handle_success(keyhalf, m);
			printf("good bye\n");
			return 0;
		}
		//if this has also failed, send KK and wait for another trace
		POSTCHECK2:
		send_result_to_fifo(KK);
		//printf("key not found with %u measure(s)\n", m);
	}
	fclose(fifo_from_wrapper);
	fclose(fifo_to_wrapper);
	return 0;
}
